==> rocm-rocBLAS-6.3.3.tar.gz: rocBLAS-rocm-6.3.3/clients/include/blas_ex/testing_dot_batched_ex.hpp <==
/* ************************************************************************
 * Copyright (C) 2018-2024 Advanced Micro Devices, Inc. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell cop-
 * ies of the Software, and to permit persons to whom the Software is furnished
 * to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IM-
 * PLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNE-
 * CTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * ************************************************************************ */

#pragma once

#include "testing_common.hpp"

template <typename Tx, typename Ty = Tx, typename Tr = Ty, typename Tex = Tr, bool CONJ = false>
void testing_dot_batched_ex_bad_arg(const Arguments& arg)
{
    auto rocblas_dot_batched_ex_fn
        = arg.api & c_API_FORTRAN
              ? (CONJ ? rocblas_dotc_batched_ex_fortran : rocblas_dot_batched_ex_fortran)
              : (CONJ ? rocblas_dotc_batched_ex : rocblas_dot_batched_ex);
    auto rocblas_dot_batched_ex_fn_64
        = arg.api & c_API_FORTRAN
              ? (CONJ ? rocblas_dotc_batched_ex_64_fortran : rocblas_dot_batched_ex_64_fortran)
              : (CONJ ? rocblas_dotc_batched_ex_64 : rocblas_dot_batched_ex_64);

    rocblas_datatype x_type         = rocblas_type2datatype<Tx>();
    rocblas_datatype y_type         = rocblas_type2datatype<Ty>();
    rocblas_datatype result_type    = rocblas_type2datatype<Tr>();
    rocblas_datatype execution_type = rocblas_type2datatype<Tex>();

    int64_t N           = 100;
    int64_t incx        = 1;
    int64_t incy        = 1;
    int64_t batch_count = 2;

    rocblas_local_handle handle{arg};

    // Allocate device memory
    device_batch_vector<Tx> dx(N, incx, batch_count);
    device_batch_vector<Ty> dy(N, incy, batch_count);
    device_vector<Tr>       d_rocblas_result(batch_count);

    // Check device memory allocation
    CHECK_DEVICE_ALLOCATION(dx.memcheck());
    CHECK_DEVICE_ALLOCATION(dy.memcheck());
    CHECK_DEVICE_ALLOCATION(d_rocblas_result.memcheck());

    CHECK_ROCBLAS_ERROR(rocblas_set_pointer_mode(handle, rocblas_pointer_mode_device));

    DAPI_EXPECT(rocblas_status_invalid_handle,
                rocblas_dot_batched_ex_fn,
                (nullptr,
                 N,
                 dx.ptr_on_device(),
                 x_type,
                 incx,
                 dy.ptr_on_device(),
                 y_type,
                 incy,
                 batch_count,
                 d_rocblas_result,
                 result_type,
                 execution_type));

    DAPI_EXPECT(rocblas_status_invalid_pointer,
                rocblas_dot_batched_ex_fn,
                (handle,
                 N,
                 nullptr,
                 x_type,
                 incx,
                 dy.ptr_on_device(),
                 y_type,
                 incy,
                 batch_count,
                 d_rocblas_result,
                 result_type,
                 execution_type));
    DAPI_EXPECT(rocblas_status_invalid_pointer,
                rocblas_dot_batched_ex_fn,
                (handle,
                 N,
                 dx.ptr_on_device(),
                 x_type,
                 incx,
                 nullptr,
                 y_type,
                 incy,
                 batch_count,
                 d_rocblas_result,
                 result_type,
                 execution_type));
    DAPI_EXPECT(rocblas_status_invalid_pointer,
                rocblas_dot_batched_ex_fn,
                (handle,
                 N,
                 dx,
                 x_type,
                 incx,
                 dy,
                 y_type,
                 incy,
                 batch_count,
                 nullptr,
                 result_type,
                 execution_type));
}

template <typename Tx, typename Ty = Tx, typename Tr = Ty, typename Tex = Tr, bool CONJ = false>
void testing_dot_batched_ex(const Arguments& arg)
{
    auto rocblas_dot_batched_ex_fn
        = arg.api & c_API_FORTRAN
              ? (CONJ ? rocblas_dotc_batched_ex_fortran : rocblas_dot_batched_ex_fortran)
              : (CONJ ? rocblas_dotc_batched_ex : rocblas_dot_batched_ex);
    auto rocblas_dot_batched_ex_fn_64
        = arg.api & c_API_FORTRAN
              ? (CONJ ? rocblas_dotc_batched_ex_64_fortran : rocblas_dot_batched_ex_64_fortran)
              : (CONJ ? rocblas_dotc_batched_ex_64 : rocblas_dot_batched_ex_64);

    rocblas_datatype x_type         = arg.a_type;
    rocblas_datatype y_type         = arg.b_type;
    rocblas_datatype result_type    = arg.c_type;
    rocblas_datatype execution_type = arg.compute_type;

    int64_t N           = arg.N;
    int64_t incx        = arg.incx;
    int64_t incy        = arg.incy;
    int64_t batch_count = arg.batch_count;

    double               rocblas_error_host   = 0;
    double               rocblas_error_device = 0;
    rocblas_local_handle handle{arg};

    // check to prevent undefined memmory allocation error
    if(N <= 0 || batch_count <= 0)
    {
        device_vector<Tr> d_rocblas_result(std::max(batch_count, int64_t(1)));
        CHECK_DEVICE_ALLOCATION(d_rocblas_result.memcheck());

        host_vector<Tr> h_rocblas_result(std::max(batch_count, int64_t(1)));
        CHECK_HIP_ERROR(h_rocblas_result.memcheck());

        CHECK_ROCBLAS_ERROR(rocblas_set_pointer_mode(handle, rocblas_pointer_mode_device));
        DAPI_CHECK(rocblas_dot_batched_ex_fn,
                   (handle,
                    N,
                    nullptr,
                    x_type,
                    incx,
                    nullptr,
                    y_type,
                    incy,
                    batch_count,
                    d_rocblas_result,
                    result_type,
                    execution_type));

        CHECK_ROCBLAS_ERROR(rocblas_set_pointer_mode(handle, rocblas_pointer_mode_host));
        DAPI_CHECK(rocblas_dot_batched_ex_fn,
                   (handle,
                    N,
                    nullptr,
                    x_type,
                    incx,
                    nullptr,
                    y_type,
                    incy,
                    batch_count,
                    h_rocblas_result,
                    result_type,
                    execution_type));

        if(batch_count > 0)
        {
            host_vector<Tr> cpu_0(batch_count);
            host_vector<Tr> gpu_0(batch_count);
            CHECK_HIP_ERROR(gpu_0.transfer_from(d_rocblas_result));
            unit_check_general<Tr>(1, 1, 1, 1, cpu_0, gpu_0, batch_count);
            unit_check_general<Tr>(1, 1, 1, 1, cpu_0, h_rocblas_result, batch_count);
        }

        return;
    }

    // Naming: `h` is in CPU (host) memory(eg hx), `d` is in GPU (device) memory (eg dx).
    // Allocate host memory
    host_batch_vector<Tx> hx(N, incx, batch_count);
    host_batch_vector<Ty> hy(N, incy, batch_count);
    host_vector<Tr>       cpu_result(batch_count);
    host_vector<Tr>       rocblas_result_host(batch_count);
    host_vector<Tr>       rocblas_result_device(batch_count);

    // Allocate device memory
    device_batch_vector<Tx> dx(N, incx, batch_count);
    device_batch_vector<Ty> dy(N, incy, batch_count);
    device_vector<Tr>       d_rocblas_result_device(batch_count);

    // Check device memory allocation
    CHECK_DEVICE_ALLOCATION(dx.memcheck());
    CHECK_DEVICE_ALLOCATION(dy.memcheck());
    CHECK_DEVICE_ALLOCATION(d_rocblas_result_device.memcheck());

    // Initialize data on host memory
    rocblas_init_vector(hx, arg, rocblas_client_alpha_sets_nan, true);
    rocblas_init_vector(hy, arg, rocblas_client_alpha_sets_nan, false);

    CHECK_HIP_ERROR(dx.transfer_from(hx));
    CHECK_HIP_ERROR(dy.transfer_from(hy));

    double gpu_time_used, cpu_time_used;

    // arg.algo indicates to force optimized x dot x kernel algorithm with equal inc
    auto  dy_ptr = (arg.algo) ? dx.ptr_on_device() : dy.ptr_on_device();
    auto& hy_ptr = (arg.algo) ? hx : hy;
    if(arg.algo)
        incy = incx;

    if(arg.unit_check || arg.norm_check)
    {
        if(arg.pointer_mode_host)
        {
            CHECK_ROCBLAS_ERROR(rocblas_set_pointer_mode(handle, rocblas_pointer_mode_host));
            DAPI_CHECK(rocblas_dot_batched_ex_fn,
                       (handle,
                        N,
                        dx.ptr_on_device(),
                        x_type,
                        incx,
                        dy_ptr,
                        y_type,
                        incy,
                        batch_count,
                        rocblas_result_host,
                        result_type,
                        execution_type));
        }

        if(arg.pointer_mode_device)
        {
            CHECK_ROCBLAS_ERROR(rocblas_set_pointer_mode(handle, rocblas_pointer_mode_device));
            handle.pre_test(arg);
            DAPI_CHECK(rocblas_dot_batched_ex_fn,
                       (handle,
                        N,
                        dx.ptr_on_device(),
                        x_type,
                        incx,
                        dy_ptr,
                        y_type,
                        incy,
                        batch_count,
                        d_rocblas_result_device,
                        result_type,
                        execution_type));
            handle.post_test(arg);

            if(arg.repeatability_check)
            {
                host_vector<Tr> rocblas_result_device_copy(batch_count);
                CHECK_HIP_ERROR(rocblas_result_device_copy.memcheck());

                CHECK_HIP_ERROR(rocblas_result_device.transfer_from(d_rocblas_result_device));

                // multi-GPU support
                int device_id, device_count;
                CHECK_HIP_ERROR(hipGetDeviceCount(&device_count));
                for(int dev_id = 0; dev_id < device_count; dev_id++)
                {
                    CHECK_HIP_ERROR(hipGetDevice(&device_id));
                    if(device_id != dev_id)
                        CHECK_HIP_ERROR(hipSetDevice(dev_id));

                    //New rocblas handle for new device
                    rocblas_local_handle handle_copy{arg};

                    //Allocate device memory in new device
                    device_batch_vector<Tx> dx_copy(N, incx, batch_count);
                    device_batch_vector<Ty> dy_copy(N, incy, batch_count);
                    device_vector<Tr>       d_rocblas_result_device_copy(batch_count);

                    // Check device memory allocation
                    CHECK_DEVICE_ALLOCATION(dx_copy.memcheck());
                    CHECK_DEVICE_ALLOCATION(dy_copy.memcheck());
                    CHECK_DEVICE_ALLOCATION(d_rocblas_result_device_copy.memcheck());

                    // copy data from CPU to device
                    CHECK_HIP_ERROR(dx_copy.transfer_from(hx));
                    CHECK_HIP_ERROR(dy_copy.transfer_from(hy));

                    auto dy_ptr_copy
                        = (arg.algo) ? dx_copy.ptr_on_device() : dy_copy.ptr_on_device();

                    CHECK_ROCBLAS_ERROR(
                        rocblas_set_pointer_mode(handle_copy, rocblas_pointer_mode_device));

                    for(int runs = 0; runs < arg.iters; runs++)
                    {
                        DAPI_CHECK(rocblas_dot_batched_ex_fn,
                                   (handle_copy,
                                    N,
                                    dx_copy.ptr_on_device(),
                                    x_type,
                                    incx,
                                    dy_ptr_copy,
                                    y_type,
                                    incy,
                                    batch_count,
                                    d_rocblas_result_device_copy,
                                    result_type,
                                    execution_type));
                        CHECK_HIP_ERROR(
                            rocblas_result_device_copy.transfer_from(d_rocblas_result_device_copy));
                        unit_check_general<Tr>(1,
                                               1,
                                               1,
                                               1,
                                               rocblas_result_device,
                                               rocblas_result_device_copy,
                                               batch_count);
                    }
                }
                return;
            }
        }

        // CPU BLAS
        cpu_time_used = get_time_us_no_sync();
        for(int b = 0; b < batch_count; ++b)
        {
            (CONJ ? ref_dotc<Tx, Tr>
                  : ref_dot<Tx, Tr>)(N, hx[b], incx, hy_ptr[b], incy, &cpu_result[b]);
        }
        cpu_time_used = get_time_us_no_sync() - cpu_time_used;

        // For large N, rocblas_half tends to diverge proportional to N
        // Tolerance is slightly greater than 1 / 1024.0
        bool near_check = arg.initialization == rocblas_initialization::hpl
                          || (std::is_same_v<Tex, rocblas_half> && N > 10000);

        if(arg.pointer_mode_host)
        {
            if(arg.unit_check)
            {
                if(near_check)
                {
                    const double tol = N * sum_error_tolerance<Tex>;
                    near_check_general<Tr>(
                        1, 1, 1, 1, cpu_result, rocblas_result_host, batch_count, tol);
                }
                else
                {
                    unit_check_general<Tr>(
                        1, 1, 1, 1, cpu_result, rocblas_result_host, batch_count);
                }
            }

            if(arg.norm_check)
            {
                for(int b = 0; b < batch_count; ++b)
                {
                    rocblas_error_host
                        += rocblas_abs((cpu_result[b] - rocblas_result_host[b]) / cpu_result[b]);
                }
            }
        }

        if(arg.pointer_mode_device)
        {
            CHECK_HIP_ERROR(rocblas_result_device.transfer_from(d_rocblas_result_device));

            if(arg.unit_check)
            {
                if(near_check)
                {
                    const double tol = N * sum_error_tolerance<Tex>;
                    near_check_general<Tr>(
                        1, 1, 1, 1, cpu_result, rocblas_result_device, batch_count, tol);
                }
                else
                {
                    unit_check_general<Tr>(
                        1, 1, 1, 1, cpu_result, rocblas_result_device, batch_count);
                }
            }

            if(arg.norm_check)
            {
                for(int b = 0; b < batch_count; ++b)
                {
                    rocblas_error_device
                        += rocblas_abs((cpu_result[b] - rocblas_result_device[b]) / cpu_result[b]);
                }
            }
        }
    }

    if(arg.timing)
    {
        int number_cold_calls = arg.cold_iters;
        int total_calls       = number_cold_calls + arg.iters;
        CHECK_ROCBLAS_ERROR(rocblas_set_pointer_mode(handle, rocblas_pointer_mode_device));

        hipStream_t stream;
        CHECK_ROCBLAS_ERROR(rocblas_get_stream(handle, &stream));

        for(int iter = 0; iter < total_calls; iter++)
        {
            if(iter == number_cold_calls)
                gpu_time_used = get_time_us_sync(stream);

            DAPI_DISPATCH(rocblas_dot_batched_ex_fn,
                          (handle,
                           N,
                           dx.ptr_on_device(),
                           x_type,
                           incx,
                           dy_ptr,
                           y_type,
                           incy,
                           batch_count,
                           d_rocblas_result_device,
                           result_type,
                           execution_type));
        }

        gpu_time_used = get_time_us_sync(stream) - gpu_time_used;

        ArgumentModel<e_N, e_incx, e_incy, e_batch_count, e_algo>{}.log_args<Tx>(
            rocblas_cout,
            arg,
            gpu_time_used,
            dot_gflop_count<CONJ, Tx>(N),
            dot_gbyte_count<Tx>(N),
            cpu_time_used,
            rocblas_error_host,
            rocblas_error_device);
    }
}

==> rocm-rocBLAS-6.4.1.tar.gz: rocBLAS-rocm-6.4.1/clients/include/blas_ex/testing_dot_batched_ex.hpp <==
/* ************************************************************************
 * Copyright (C) 2018-2025 Advanced Micro Devices, Inc. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell cop-
 * ies of the Software, and to permit persons to whom the Software is furnished
 * to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IM-
 * PLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNE-
 * CTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * ************************************************************************ */

#pragma once

#include "testing_common.hpp"

template <typename Tx, typename Ty = Tx, typename Tr = Ty, typename Tex = Tr, bool CONJ = false>
void testing_dot_batched_ex_bad_arg(const Arguments& arg)
{
    auto rocblas_dot_batched_ex_fn
        = arg.api & c_API_FORTRAN
              ? (CONJ ? rocblas_dotc_batched_ex_fortran : rocblas_dot_batched_ex_fortran)
              : (CONJ ? rocblas_dotc_batched_ex : rocblas_dot_batched_ex);
    auto rocblas_dot_batched_ex_fn_64
        = arg.api & c_API_FORTRAN
              ? (CONJ ? rocblas_dotc_batched_ex_64_fortran : rocblas_dot_batched_ex_64_fortran)
              : (CONJ ? rocblas_dotc_batched_ex_64 : rocblas_dot_batched_ex_64);

    rocblas_datatype x_type         = rocblas_type2datatype<Tx>();
    rocblas_datatype y_type         = rocblas_type2datatype<Ty>();
    rocblas_datatype result_type    = rocblas_type2datatype<Tr>();
    rocblas_datatype execution_type = rocblas_type2datatype<Tex>();

    int64_t N           = 100;
    int64_t incx        = 1;
    int64_t incy        = 1;
    int64_t batch_count = 2;

    rocblas_local_handle handle{arg};

    // Allocate device memory
    DEVICE_MEMCHECK(device_batch_vector<Tx>, dx, (N, incx, batch_count));
    DEVICE_MEMCHECK(device_batch_vector<Ty>, dy, (N, incy, batch_count));
    DEVICE_MEMCHECK(device_vector<Tr>, d_rocblas_result, (batch_count));

    CHECK_ROCBLAS_ERROR(rocblas_set_pointer_mode(handle, rocblas_pointer_mode_device));

    DAPI_EXPECT(rocblas_status_invalid_handle,
                rocblas_dot_batched_ex_fn,
                (nullptr,
                 N,
                 dx.ptr_on_device(),
                 x_type,
                 incx,
                 dy.ptr_on_device(),
                 y_type,
                 incy,
                 batch_count,
                 d_rocblas_result,
                 result_type,
                 execution_type));

    DAPI_EXPECT(rocblas_status_invalid_pointer,
                rocblas_dot_batched_ex_fn,
                (handle,
                 N,
                 nullptr,
                 x_type,
                 incx,
                 dy.ptr_on_device(),
                 y_type,
                 incy,
                 batch_count,
                 d_rocblas_result,
                 result_type,
                 execution_type));
    DAPI_EXPECT(rocblas_status_invalid_pointer,
                rocblas_dot_batched_ex_fn,
                (handle,
                 N,
                 dx.ptr_on_device(),
                 x_type,
                 incx,
                 nullptr,
                 y_type,
                 incy,
                 batch_count,
                 d_rocblas_result,
                 result_type,
                 execution_type));
    DAPI_EXPECT(rocblas_status_invalid_pointer,
                rocblas_dot_batched_ex_fn,
                (handle,
                 N,
                 dx,
                 x_type,
                 incx,
                 dy,
                 y_type,
                 incy,
                 batch_count,
                 nullptr,
                 result_type,
                 execution_type));
}

template <typename Tx, typename Ty = Tx, typename Tr = Ty, typename Tex = Tr, bool CONJ = false>
void testing_dot_batched_ex(const Arguments& arg)
{
    auto rocblas_dot_batched_ex_fn
        = arg.api & c_API_FORTRAN
              ? (CONJ ? rocblas_dotc_batched_ex_fortran : rocblas_dot_batched_ex_fortran)
              : (CONJ ? rocblas_dotc_batched_ex : rocblas_dot_batched_ex);
    auto rocblas_dot_batched_ex_fn_64
        = arg.api & c_API_FORTRAN
              ? (CONJ ? rocblas_dotc_batched_ex_64_fortran : rocblas_dot_batched_ex_64_fortran)
              : (CONJ ? rocblas_dotc_batched_ex_64 : rocblas_dot_batched_ex_64);

    rocblas_datatype x_type         = arg.a_type;
    rocblas_datatype y_type         = arg.b_type;
    rocblas_datatype result_type    = arg.c_type;
    rocblas_datatype execution_type = arg.compute_type;

    int64_t N           = arg.N;
    int64_t incx        = arg.incx;
    int64_t incy        = arg.incy;
    int64_t batch_count = arg.batch_count;

    double               rocblas_error_host   = 0;
    double               rocblas_error_device = 0;
    rocblas_local_handle handle{arg};

    // check to prevent undefined memmory allocation error
    if(N <= 0 || batch_count <= 0)
    {
        DEVICE_MEMCHECK(device_vector<Tr>, d_rocblas_result, (std::max(batch_count, int64_t(1))));

        HOST_MEMCHECK(host_vector<Tr>, h_rocblas_result, (std::max(batch_count, int64_t(1))));

        CHECK_ROCBLAS_ERROR(rocblas_set_pointer_mode(handle, rocblas_pointer_mode_device));
        DAPI_CHECK(rocblas_dot_batched_ex_fn,
                   (handle,
                    N,
                    nullptr,
                    x_type,
                    incx,
                    nullptr,
                    y_type,
                    incy,
                    batch_count,
                    d_rocblas_result,
                    result_type,
                    execution_type));

        CHECK_ROCBLAS_ERROR(rocblas_set_pointer_mode(handle, rocblas_pointer_mode_host));
        DAPI_CHECK(rocblas_dot_batched_ex_fn,
                   (handle,
                    N,
                    nullptr,
                    x_type,
                    incx,
                    nullptr,
                    y_type,
                    incy,
                    batch_count,
                    h_rocblas_result,
                    result_type,
                    execution_type));

        if(batch_count > 0)
        {
            HOST_MEMCHECK(host_vector<Tr>, cpu_0, (batch_count));
            HOST_MEMCHECK(host_vector<Tr>, gpu_0, (batch_count));
            CHECK_HIP_ERROR(gpu_0.transfer_from(d_rocblas_result));
            unit_check_general<Tr>(1, 1, 1, 1, cpu_0, gpu_0, batch_count);
            unit_check_general<Tr>(1, 1, 1, 1, cpu_0, h_rocblas_result, batch_count);
        }

        return;
    }

    // Naming: `h` is in CPU (host) memory(eg hx), `d` is in GPU (device) memory (eg dx).
    // Allocate host memory
    HOST_MEMCHECK(host_batch_vector<Tx>, hx, (N, incx, batch_count));
    HOST_MEMCHECK(host_batch_vector<Ty>, hy, (N, incy, batch_count));
    HOST_MEMCHECK(host_vector<Tr>, cpu_result, (batch_count));
    HOST_MEMCHECK(host_vector<Tr>, rocblas_result, (batch_count));

    // Allocate device memory
    DEVICE_MEMCHECK(device_batch_vector<Tx>, dx, (N, incx, batch_count));
    DEVICE_MEMCHECK(device_batch_vector<Ty>, dy, (N, incy, batch_count));
    DEVICE_MEMCHECK(device_vector<Tr>, d_rocblas_result_device, (batch_count));

    // Initialize data on host memory
    rocblas_init_vector(hx, arg, rocblas_client_alpha_sets_nan, true);
    rocblas_init_vector(hy, arg, rocblas_client_alpha_sets_nan, false);

    CHECK_HIP_ERROR(dx.transfer_from(hx));
    CHECK_HIP_ERROR(dy.transfer_from(hy));

    double gpu_time_used, cpu_time_used;

    // arg.algo indicates to force optimized x dot x kernel algorithm with equal inc
    auto  dy_ptr = (arg.algo) ? dx.ptr_on_device() : dy.ptr_on_device();
    auto& hy_ptr = (arg.algo) ? hx : hy;
    if(arg.algo)
        incy = incx;

    if(arg.unit_check || arg.norm_check)
    {
        if(arg.pointer_mode_host)
        {
            CHECK_ROCBLAS_ERROR(rocblas_set_pointer_mode(handle, rocblas_pointer_mode_host));
            DAPI_CHECK(rocblas_dot_batched_ex_fn,
                       (handle,
                        N,
                        dx.ptr_on_device(),
                        x_type,
                        incx,
                        dy_ptr,
                        y_type,
                        incy,
                        batch_count,
                        rocblas_result,
                        result_type,
                        execution_type));
        }

        if(arg.pointer_mode_device)
        {
            CHECK_ROCBLAS_ERROR(rocblas_set_pointer_mode(handle, rocblas_pointer_mode_device));
            handle.pre_test(arg);
            DAPI_CHECK(rocblas_dot_batched_ex_fn,
                       (handle,
                        N,
                        dx.ptr_on_device(),
                        x_type,
                        incx,
                        dy_ptr,
                        y_type,
                        incy,
                        batch_count,
                        d_rocblas_result_device,
                        result_type,
                        execution_type));
            handle.post_test(arg);

            if(arg.repeatability_check)
            {
                HOST_MEMCHECK(host_vector<Tr>, rocblas_result_device_copy, (batch_count));

                CHECK_HIP_ERROR(rocblas_result.transfer_from(d_rocblas_result_device));

                // multi-GPU support
                int device_id, device_count;
                CHECK_HIP_ERROR(limit_device_count(device_count, (int)arg.devices));

                for(int dev_id = 0; dev_id < device_count; dev_id++)
                {
                    CHECK_HIP_ERROR(hipGetDevice(&device_id));
                    if(device_id != dev_id)
                        CHECK_HIP_ERROR(hipSetDevice(dev_id));

                    //New rocblas handle for new device
                    rocblas_local_handle handle_copy{arg};

                    //Allocate device memory in new device
                    DEVICE_MEMCHECK(device_batch_vector<Tx>, dx_copy, (N, incx, batch_count));
                    DEVICE_MEMCHECK(device_batch_vector<Ty>, dy_copy, (N, incy, batch_count));
                    DEVICE_MEMCHECK(device_vector<Tr>, d_rocblas_result_device_copy, (batch_count));

                    // copy data from CPU to device
                    CHECK_HIP_ERROR(dx_copy.transfer_from(hx));
                    CHECK_HIP_ERROR(dy_copy.transfer_from(hy));

                    auto dy_ptr_copy
                        = (arg.algo) ? dx_copy.ptr_on_device() : dy_copy.ptr_on_device();

                    CHECK_ROCBLAS_ERROR(
                        rocblas_set_pointer_mode(handle_copy, rocblas_pointer_mode_device));

                    for(int runs = 0; runs < arg.iters; runs++)
                    {
                        DAPI_CHECK(rocblas_dot_batched_ex_fn,
                                   (handle_copy,
                                    N,
                                    dx_copy.ptr_on_device(),
                                    x_type,
                                    incx,
                                    dy_ptr_copy,
                                    y_type,
                                    incy,
                                    batch_count,
                                    d_rocblas_result_device_copy,
                                    result_type,
                                    execution_type));
                        CHECK_HIP_ERROR(
                            rocblas_result_device_copy.transfer_from(d_rocblas_result_device_copy));
                        unit_check_general<Tr>(
                            1, 1, 1, 1, rocblas_result, rocblas_result_device_copy, batch_count);
                    }
                }
                return;
            }
        }

        // CPU BLAS
        cpu_time_used = get_time_us_no_sync();
        for(int64_t b = 0; b < batch_count; ++b)
        {
            (CONJ ? ref_dotc<Tx, Tr>
                  : ref_dot<Tx, Tr>)(N, hx[b], incx, hy_ptr[b], incy, &cpu_result[b]);
        }
        cpu_time_used = get_time_us_no_sync() - cpu_time_used;

        auto compare_to_gold = [&] {
            bool near_check = arg.initialization == rocblas_initialization::hpl
                              || (sizeof(Tr) == 2 && N > 10000);

            if(arg.unit_check)
            {
                if(near_check)
                {
                    for(int64_t b = 0; b < batch_count; ++b)
                    {
                        double abs_error = dot_near_tolerance<Tex, Tx, Tr>(
                            rocblas_handle(handle)->getArchMajor(), N, cpu_result[b]);
                        near_check_general<Tr>(
                            1, 1, 1, 1, cpu_result + b, rocblas_result + b, 1, abs_error);
                    }
                }
                else
                {
                    unit_check_general<Tr>(1, 1, 1, 1, cpu_result, rocblas_result, batch_count);
                }
            }

            double error = 0.0;
            if(arg.norm_check)
            {
                error = rocblas_abs((cpu_result[0] - rocblas_result[0]) / cpu_result[0]);
            }
            return error;
        };

        if(arg.pointer_mode_host)
        {
            rocblas_error_host = compare_to_gold();
        }

        if(arg.pointer_mode_device)
        {
            CHECK_HIP_ERROR(rocblas_result.transfer_from(d_rocblas_result_device));
            rocblas_error_device = compare_to_gold();
        }
    }

    if(arg.timing)
    {
        int number_cold_calls = arg.cold_iters;
        int total_calls       = number_cold_calls + arg.iters;
        CHECK_ROCBLAS_ERROR(rocblas_set_pointer_mode(handle, rocblas_pointer_mode_device));

        hipStream_t stream;
        CHECK_ROCBLAS_ERROR(rocblas_get_stream(handle, &stream));

        for(int iter = 0; iter < total_calls; iter++)
        {
            if(iter == number_cold_calls)
                gpu_time_used = get_time_us_sync(stream);

            DAPI_DISPATCH(rocblas_dot_batched_ex_fn,
                          (handle,
                           N,
                           dx.ptr_on_device(),
                           x_type,
                           incx,
                           dy_ptr,
                           y_type,
                           incy,
                           batch_count,
                           d_rocblas_result_device,
                           result_type,
                           execution_type));
        }

        gpu_time_used = get_time_us_sync(stream) - gpu_time_used;

        ArgumentModel<e_N, e_incx, e_incy, e_batch_count, e_algo>{}.log_args<Tx>(
            rocblas_cout,
            arg,
            gpu_time_used,
            dot_gflop_count<CONJ, Tx>(N),
            dot_gbyte_count<Tx>(N),
            cpu_time_used,
            rocblas_error_host,
            rocblas_error_device);
    }
}
